@@ -105,12 +105,8 @@
  * application.
  */
 
-#include "ld_syscall.h"
-#include "linuxelf.h"
-#include "ld_hash.h"
-#include "ld_string.h"
-#include "dlfcn.h"
-#include "../config.h"
+#include "ldso.h"
+
 
 #define ALLOW_ZERO_PLTGOT
 
@@ -129,16 +125,16 @@
 char *_dl_library_path = 0;		/* Where we look for libraries */
 char *_dl_preload = 0;			/* Things to be loaded before the libs. */
 char *_dl_ldsopath = 0;
-static char *_dl_not_lazy = 0;
+static int _dl_be_lazy = RTLD_LAZY;
 #ifdef __SUPPORT_LD_DEBUG__
-static char *_dl_debug  = 0;
-static char *_dl_debug_symbols = 0;
-static char *_dl_debug_move    = 0;
-static char *_dl_debug_reloc   = 0;
-static char *_dl_debug_detail  = 0;
-static char *_dl_debug_nofixups  = 0;
-static char *_dl_debug_bindings  = 0;
-static int   _dl_debug_file = 2;
+char *_dl_debug  = 0;
+char *_dl_debug_symbols = 0;
+char *_dl_debug_move    = 0;
+char *_dl_debug_reloc   = 0;
+char *_dl_debug_detail  = 0;
+char *_dl_debug_nofixups  = 0;
+char *_dl_debug_bindings  = 0;
+int   _dl_debug_file = 2;
 #else
 #define _dl_debug_file 2
 #endif
@@ -146,19 +142,18 @@ static char *_dl_malloc_addr, *_dl_mmap_zero;
 
 static char *_dl_trace_loaded_objects = 0;
 static int (*_dl_elf_main) (int, char **, char **);
-static int (*_dl_elf_init) (void);
 struct r_debug *_dl_debug_addr = NULL;
 unsigned long *_dl_brkp;
 unsigned long *_dl_envp;
-int _dl_fixup(struct elf_resolve *tpnt);
+int _dl_fixup(struct elf_resolve *tpnt, int lazy);
 void _dl_debug_state(void);
 char *_dl_get_last_path_component(char *path);
 static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *app_tpnt, 
 		unsigned long load_addr, unsigned long *hash_addr, Elf32_auxv_t auxvt[AT_EGID + 1], 
 		char **envp, struct r_debug *debug_addr);
-#include "boot1_arch.h"
-#include "ldso.h"				/* Pull in the name of ld.so */
 
+#include "boot1_arch.h"
+#include "_dl_progname.h"				/* Pull in the value of _dl_progname */
 
 /* When we enter this piece of code, the program stack looks like this:
         argc            argument counter (integer)
@@ -202,7 +197,7 @@ LD_BOOT(unsigned long args)
 	unsigned long *got;
 	unsigned long *aux_dat;
 	int goof = 0;
-	elfhdr *header;
+	ElfW(Ehdr) *header;
 	struct elf_resolve *tpnt;
 	struct elf_resolve *app_tpnt;
 	Elf32_auxv_t auxvt[AT_EGID + 1];
@@ -222,7 +217,7 @@ LD_BOOT(unsigned long args)
 	/* First obtain the information on the stack that tells us more about
 	   what binary is loaded, where it is loaded, etc, etc */
 	GET_ARGV(aux_dat, args);
-#if defined (__arm__) || defined (__mips__)
+#if defined (__arm__) || defined (__mips__) || defined (__cris__)
 	aux_dat += 1;
 #endif
 	argc = *(aux_dat - 1);
@@ -253,7 +248,7 @@ LD_BOOT(unsigned long args)
 	/* locate the ELF header.   We need this done as soon as possible 
 	 * (esp since SEND_STDERR() needs this on some platforms... */
 	load_addr = auxvt[AT_BASE].a_un.a_val;
-	header = (elfhdr *) auxvt[AT_BASE].a_un.a_ptr;
+	header = (ElfW(Ehdr) *) auxvt[AT_BASE].a_un.a_ptr;
 
 	/* Check the ELF header to make sure everything looks ok.  */
 	if (!header || header->e_ident[EI_CLASS] != ELFCLASS32 ||
@@ -292,7 +287,7 @@ LD_BOOT(unsigned long args)
   __asm__("\tbl _GLOBAL_OFFSET_TABLE_-4@local\n\t":"=l"(got));
 #elif defined(__mips__)
   __asm__("\tmove %0, $28\n\tsubu %0,%0,0x7ff0\n\t":"=r"(got));
-#elif defined(__sh__)
+#elif defined(__sh__) && !defined(__SH5__)
   __asm__(
 "       mov.l    1f, %0\n"
 "       mova     1f, r0\n"
@@ -364,7 +359,7 @@ LD_BOOT(unsigned long args)
 
 	/* Call mmap to get a page of writable memory that can be used 
 	 * for _dl_malloc throughout the shared lib loader. */
-	mmap_zero = malloc_buffer = _dl_mmap((void *) 0, 4096, 
+	mmap_zero = malloc_buffer = _dl_mmap((void *) 0, PAGE_SIZE, 
 			PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
 	if (_dl_mmap_check_error(mmap_zero)) {
 		SEND_STDERR("dl_boot: mmap of a spare page failed!\n");
@@ -376,6 +371,28 @@ LD_BOOT(unsigned long args)
 	app_tpnt = LD_MALLOC(sizeof(struct elf_resolve));
 	_dl_memset(app_tpnt, 0, sizeof(struct elf_resolve));
 
+#ifdef __UCLIBC_PIE_SUPPORT__
+	/* Find the runtime load address of the main executable, this may be
+         * different from what the ELF header says for ET_DYN/PIE executables.
+	 */
+	{
+		ElfW(Phdr) *ppnt;
+		int i;
+
+		ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
+		for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++)
+			if (ppnt->p_type == PT_PHDR) {
+				app_tpnt->loadaddr = (ElfW(Addr)) (auxvt[AT_PHDR].a_un.a_val - ppnt->p_vaddr);
+				break;
+			}
+	}
+
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+	SEND_STDERR("app_tpnt->loadaddr=");
+	SEND_ADDRESS_STDERR(app_tpnt->loadaddr, 1);
+#endif
+#endif
+
 	/*
 	 * This is used by gdb to locate the chain of shared libraries that are currently loaded.
 	 */
@@ -406,13 +423,17 @@ LD_BOOT(unsigned long args)
 	}
 
 	{
-		elf_phdr *ppnt;
+		ElfW(Phdr) *ppnt;
 		int i;
 
-		ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+		ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 		for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++)
 			if (ppnt->p_type == PT_DYNAMIC) {
+#ifndef __UCLIBC_PIE_SUPPORT__
 				dpnt = (Elf32_Dyn *) ppnt->p_vaddr;
+#else
+				dpnt = (Elf32_Dyn *) (ppnt->p_vaddr + app_tpnt->loadaddr);
+#endif
 				while (dpnt->d_tag) {
 #if defined(__mips__)
 					if (dpnt->d_tag == DT_MIPS_GOTSYM)
@@ -424,18 +445,28 @@ LD_BOOT(unsigned long args)
 					if (dpnt->d_tag == DT_MIPS_SYMTABNO)
 						app_tpnt->mips_symtabno =
 							(unsigned long) dpnt->d_un.d_val;
-#endif
 					if (dpnt->d_tag > DT_JMPREL) {
 						dpnt++;
 						continue;
 					}
 					app_tpnt->dynamic_info[dpnt->d_tag] = dpnt->d_un.d_val;
-#if !defined(__mips__)
+
+#warning "Debugging threads on mips won't work till someone fixes this..."
+#if 0
 					if (dpnt->d_tag == DT_DEBUG) {
 						dpnt->d_un.d_val = (unsigned long) debug_addr;
 					}
+#endif
+
 #else
-#warning "Debugging threads on mips won't work till someone fixes this..."
+					if (dpnt->d_tag > DT_JMPREL) {
+						dpnt++;
+						continue;
+					}
+					app_tpnt->dynamic_info[dpnt->d_tag] = dpnt->d_un.d_val;
+					if (dpnt->d_tag == DT_DEBUG) {
+						dpnt->d_un.d_val = (unsigned long) debug_addr;
+					}
 #endif
 					if (dpnt->d_tag == DT_TEXTREL)
 						app_tpnt->dynamic_info[DT_TEXTREL] = 1;
@@ -467,7 +498,7 @@ LD_BOOT(unsigned long args)
 	   protection back again once we are done */
 
 	{
-		elf_phdr *ppnt;
+		ElfW(Phdr) *ppnt;
 		int i;
 
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
@@ -476,8 +507,8 @@ LD_BOOT(unsigned long args)
 
 		/* First cover the shared library/dynamic linker. */
 		if (tpnt->dynamic_info[DT_TEXTREL]) {
-			header = (elfhdr *) auxvt[AT_BASE].a_un.a_ptr;
-			ppnt = (elf_phdr *) ((int)auxvt[AT_BASE].a_un.a_ptr + 
+			header = (ElfW(Ehdr) *) auxvt[AT_BASE].a_un.a_ptr;
+			ppnt = (ElfW(Phdr) *) ((int)auxvt[AT_BASE].a_un.a_ptr + 
 					header->e_phoff);
 			for (i = 0; i < header->e_phnum; i++, ppnt++) {
 				if (ppnt->p_type == PT_LOAD && !(ppnt->p_flags & PF_W)) {
@@ -493,11 +524,16 @@ LD_BOOT(unsigned long args)
 #endif
 		/* Now cover the application program. */
 		if (app_tpnt->dynamic_info[DT_TEXTREL]) {
-			ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+			ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 			for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++) {
 				if (ppnt->p_type == PT_LOAD && !(ppnt->p_flags & PF_W))
+#ifndef __UCLIBC_PIE_SUPPORT__
 					_dl_mprotect((void *) (ppnt->p_vaddr & PAGE_ALIGN),
 								 (ppnt->p_vaddr & ADDR_ALIGN) +
+#else
+					_dl_mprotect((void *) ((ppnt->p_vaddr + app_tpnt->loadaddr) & PAGE_ALIGN),
+								 ((ppnt->p_vaddr + app_tpnt->loadaddr) & ADDR_ALIGN) +
+#endif
 								 (unsigned long) ppnt->p_filesz,
 								 PROT_READ | PROT_WRITE | PROT_EXEC);
 			}
@@ -578,11 +614,13 @@ LD_BOOT(unsigned long args)
 				SEND_STDERR(strtab + symtab[symtab_index].st_name);
 				SEND_STDERR("\n");
 #endif  
+				PERFORM_BOOTSTRAP_RELOC(rpnt, reloc_addr, symbol_addr, load_addr, &symtab[symtab_index]);
+			} else {
+				/*
+				 * Use this machine-specific macro to perform the actual relocation.
+				 */
+				PERFORM_BOOTSTRAP_RELOC(rpnt, reloc_addr, symbol_addr, load_addr, NULL);
 			}
-			/*
-			 * Use this machine-specific macro to perform the actual relocation.
-			 */
-			PERFORM_BOOTSTRAP_RELOC(rpnt, reloc_addr, symbol_addr, load_addr);
 		}
 	}
 
@@ -647,7 +685,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 		unsigned long load_addr, unsigned long *hash_addr, Elf32_auxv_t auxvt[AT_EGID + 1], 
 		char **envp, struct r_debug *debug_addr)
 {
-	elf_phdr *ppnt;
+	ElfW(Phdr) *ppnt;
 	char *lpntstr;
 	int i, _dl_secure, goof = 0;
 	struct dyn_elf *rpnt;
@@ -669,7 +707,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	tpnt->next = 0;
 	tpnt->libname = 0;
 	tpnt->libtype = program_interpreter;
-	tpnt->loadaddr = (char *) load_addr;
+	tpnt->loadaddr = (ElfW(Addr)) load_addr;
 
 #ifdef ALLOW_ZERO_PLTGOT
 	if (tpnt->dynamic_info[DT_PLTGOT])
@@ -685,30 +723,17 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	   and load them properly. */
 
 	{
-		elfhdr *epnt;
-		elf_phdr *myppnt;
+		ElfW(Ehdr) *epnt;
+		ElfW(Phdr) *myppnt;
 		int j;
 
-		epnt = (elfhdr *) auxvt[AT_BASE].a_un.a_ptr;
+		epnt = (ElfW(Ehdr) *) auxvt[AT_BASE].a_un.a_ptr;
 		tpnt->n_phent = epnt->e_phnum;
-		tpnt->ppnt = myppnt = (elf_phdr *) (load_addr + epnt->e_phoff);
+		tpnt->ppnt = myppnt = (ElfW(Phdr) *) (load_addr + epnt->e_phoff);
 		for (j = 0; j < epnt->e_phnum; j++, myppnt++) {
 			if (myppnt->p_type == PT_DYNAMIC) {
-				tpnt->dynamic_addr = myppnt->p_vaddr + load_addr;
-#if defined(__mips__)
-				{
-					int k = 1;
-					Elf32_Dyn *dpnt = (Elf32_Dyn *) tpnt->dynamic_addr;
-
-					while(dpnt->d_tag) {
-						dpnt++;
-						k++;
-					}
-					tpnt->dynamic_size = k * sizeof(Elf32_Dyn);
-				}
-#else
+				tpnt->dynamic_addr = (ElfW(Dyn) *)myppnt->p_vaddr + load_addr;
 				tpnt->dynamic_size = myppnt->p_filesz;
-#endif
 			}
 		}
 	}
@@ -720,11 +745,16 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	   and figure out which libraries are supposed to be called.  Until
 	   we have this list, we will not be completely ready for dynamic linking */
 
-	ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+	ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 	for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++) {
 		if (ppnt->p_type == PT_LOAD) {
+#ifndef __UCLIBC_PIE_SUPPORT__
 			if (ppnt->p_vaddr + ppnt->p_memsz > brk_addr)
 				brk_addr = ppnt->p_vaddr + ppnt->p_memsz;
+#else
+			if (ppnt->p_vaddr + app_tpnt->loadaddr + ppnt->p_memsz > brk_addr)
+				brk_addr = ppnt->p_vaddr + app_tpnt->loadaddr + ppnt->p_memsz;
+#endif
 		}
 		if (ppnt->p_type == PT_DYNAMIC) {
 #ifndef ALLOW_ZERO_PLTGOT
@@ -733,32 +763,26 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 				continue;
 #endif
 			/* OK, we have what we need - slip this one into the list. */
-#if defined(__mips__)
-			{
-				int i = 1;
-				Elf32_Dyn *dpnt = (Elf32_Dyn *) tpnt->dynamic_addr;
-
-				while(dpnt->d_tag) {
-					dpnt++;
-					i++;
-				}
-				app_tpnt = _dl_add_elf_hash_table("", 0, 
-					app_tpnt->dynamic_info, ppnt->p_vaddr,
-					(i * sizeof(Elf32_Dyn)));
-			}
-#else
+#ifndef __UCLIBC_PIE_SUPPORT__
 			app_tpnt = _dl_add_elf_hash_table("", 0, 
 					app_tpnt->dynamic_info, ppnt->p_vaddr, ppnt->p_filesz);
+#else
+			app_tpnt = _dl_add_elf_hash_table("", (char *)app_tpnt->loadaddr,
+					app_tpnt->dynamic_info, ppnt->p_vaddr + app_tpnt->loadaddr, ppnt->p_filesz);
 #endif
 			_dl_loaded_modules->libtype = elf_executable;
-			_dl_loaded_modules->ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+			_dl_loaded_modules->ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 			_dl_loaded_modules->n_phent = auxvt[AT_PHNUM].a_un.a_val;
 			_dl_symbol_tables = rpnt = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
 			_dl_memset(rpnt, 0, sizeof(struct dyn_elf));
 			rpnt->dyn = _dl_loaded_modules;
 			app_tpnt->usage_count++;
 			app_tpnt->symbol_scope = _dl_symbol_tables;
+#ifndef __UCLIBC_PIE_SUPPORT__
 			lpnt = (unsigned long *) (app_tpnt->dynamic_info[DT_PLTGOT]);
+#else
+			lpnt = (unsigned long *) (app_tpnt->dynamic_info[DT_PLTGOT] + app_tpnt->loadaddr);
+#endif
 #ifdef ALLOW_ZERO_PLTGOT
 			if (lpnt)
 #endif
@@ -805,7 +829,8 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	/* Now we need to figure out what kind of options are selected.
 	   Note that for SUID programs we ignore the settings in LD_LIBRARY_PATH */
 	{
-		_dl_not_lazy = _dl_getenv("LD_BIND_NOW", envp);
+		if (_dl_getenv("LD_BIND_NOW", envp))
+			_dl_be_lazy = 0;
 
 		if ((auxvt[AT_UID].a_un.a_val == -1 && _dl_suid_ok()) ||
 				(auxvt[AT_UID].a_un.a_val != -1 && 
@@ -905,9 +930,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	   libraries that should be loaded, and insert them on the list in the
 	   correct order. */
 
-#ifdef USE_CACHE
 	_dl_map_cache();
-#endif
 
 
 	if (_dl_preload) 
@@ -926,6 +949,14 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 			*str2 = '\0';
 			if (!_dl_secure || _dl_strchr(str, '/') == NULL) 
 			{
+				if ((tpnt1 = _dl_check_if_named_library_is_loaded(str))) 
+				{
+					continue;
+				}
+#if defined (__SUPPORT_LD_DEBUG__)
+				if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfile='%s';  needed by '%s'\n", 
+						str, _dl_progname);
+#endif
 				tpnt1 = _dl_load_shared_library(_dl_secure, &rpnt, NULL, str);
 				if (!tpnt1) {
 #ifdef __LDSO_LDD_SUPPORT__
@@ -1002,6 +1033,14 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 						c = *cp;
 						*cp = '\0';
 
+						if ((tpnt1 = _dl_check_if_named_library_is_loaded(cp2))) 
+						{
+							continue;
+						}
+#if defined (__SUPPORT_LD_DEBUG__)
+						if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfile='%s';  needed by '%s'\n", 
+								cp2, _dl_progname);
+#endif
 						tpnt1 = _dl_load_shared_library(0, &rpnt, NULL, cp2);
 						if (!tpnt1) {
 #ifdef __LDSO_LDD_SUPPORT__
@@ -1038,67 +1077,32 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	}
 #endif
 
-	for (tcurr = _dl_loaded_modules; tcurr; tcurr = tcurr->next) {
+	for (tcurr = _dl_loaded_modules; tcurr; tcurr = tcurr->next) 
+	{
 		Elf32_Dyn *dpnt;
-		for (dpnt = (Elf32_Dyn *) tcurr->dynamic_addr; dpnt->d_tag;
-				dpnt++) {
-			if (dpnt->d_tag == DT_NEEDED) {
-				lpntstr = tcurr->loadaddr + tcurr->dynamic_info[DT_STRTAB] +
-					dpnt->d_un.d_val;
-				if (_dl_strcmp(lpntstr, "libc.so.6") == 0) {
-					char *name, *msg;
-					name = tcurr->libname;
-					while(*name == '/')
-						name++;
-					if (_dl_trace_loaded_objects) {
-						msg = "WARNING"; 
-					} else {
-						msg = "ERROR"; 
-					}
-					_dl_dprintf(2, "\t%s: %s is linked with GNU libc!\n", msg, --name);
-					/* If all we are doing is ldd, then we don't need to freak out... */
-					if (_dl_trace_loaded_objects) {
-						continue;
-					}
-					/* Time to freak out.  Make sure glibc linked libraries are not loaded */
-					_dl_exit(150);
-				}
-				if (tpnt && _dl_strcmp(lpntstr, _dl_get_last_path_component(tpnt->libname)) == 0) 
-				{
-					struct elf_resolve *ttmp;
+		for (dpnt = (Elf32_Dyn *) tcurr->dynamic_addr; dpnt->d_tag; dpnt++) 
+		{
+			if (dpnt->d_tag == DT_NEEDED) 
+			{
+				char *name;
+				lpntstr = (char*) (tcurr->loadaddr + tcurr->dynamic_info[DT_STRTAB] + dpnt->d_un.d_val);
+				name = _dl_get_last_path_component(lpntstr);
 
-#ifdef __LDSO_LDD_SUPPORT__
-					if (_dl_trace_loaded_objects && tpnt->usage_count==1) {
-						char *name;
-						name = tpnt->libname;
-						while(*name == '/')
-							name++;
-						_dl_dprintf(1, "\t%s => %s (%x)\n", 
-								lpntstr, --name, (unsigned) tpnt->loadaddr);
-					}
-#endif
-					ttmp = _dl_loaded_modules;
-					while (ttmp->next)
-						ttmp = ttmp->next;
-					ttmp->next = tpnt;
-					tpnt->prev = ttmp;
-					tpnt->next = NULL;
-					rpnt->next = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
-					_dl_memset(rpnt->next, 0, sizeof(struct dyn_elf));
-					rpnt->next->prev = rpnt;
-					rpnt = rpnt->next;
-					rpnt->dyn = tpnt;
-					tpnt->usage_count++;
-					tpnt->symbol_scope = _dl_symbol_tables;
-					tpnt = NULL;
+				if ((tpnt1 = _dl_check_if_named_library_is_loaded(name))) 
+				{
 					continue;
 				}
+#if defined (__SUPPORT_LD_DEBUG__)
+				if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfile='%s';  needed by '%s'\n", 
+						lpntstr, _dl_progname);
+#endif
 				if (!(tpnt1 = _dl_load_shared_library(0, &rpnt, tcurr, lpntstr)))
 				{
 #ifdef __LDSO_LDD_SUPPORT__
-					if (_dl_trace_loaded_objects)
+					if (_dl_trace_loaded_objects) {
 						_dl_dprintf(1, "\t%s => not found\n", lpntstr);
-					else 
+						continue;
+					} else 
 #endif
 					{
 						_dl_dprintf(2, "%s: can't load library '%s'\n", _dl_progname, lpntstr);
@@ -1110,11 +1114,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 #endif
 #ifdef __LDSO_LDD_SUPPORT__
 					if (_dl_trace_loaded_objects && tpnt1->usage_count==1) {
-						char *name;
-						name = tpnt1->libname;
-						while(*name == '/')
-							name++;
-						_dl_dprintf(1, "\t%s => %s (%x)\n", lpntstr, --name, 
+						_dl_dprintf(1, "\t%s => %s (%x)\n", lpntstr, tpnt1->libname, 
 								(unsigned) tpnt1->loadaddr);
 					}
 #endif
@@ -1124,9 +1124,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	}
 
 
-#ifdef USE_CACHE
 	_dl_unmap_cache();
-#endif
 
 	/*
 	 * If the program interpreter is not in the module chain, add it.  This will
@@ -1187,7 +1185,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	 * Now we go through and look for REL and RELA records that indicate fixups
 	 * to the GOT tables.  We need to do this in reverse order so that COPY
 	 * directives work correctly */
-	goof = _dl_loaded_modules ? _dl_fixup(_dl_loaded_modules) : 0;
+	goof = _dl_loaded_modules ? _dl_fixup(_dl_loaded_modules, _dl_be_lazy) : 0;
 
 
 	/* Some flavors of SVr4 do not generate the R_*_COPY directive,
@@ -1222,7 +1220,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 #ifndef FORCE_SHAREABLE_TEXT_SEGMENTS
 	{
 		unsigned int j;
-		elf_phdr *myppnt;
+		ElfW(Phdr) *myppnt;
 
 		/* We had to set the protections of all pages to R/W for dynamic linking.
 		   Set text pages back to R/O */
@@ -1268,15 +1266,17 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 		tpnt->init_flag |= INIT_FUNCS_CALLED;
 
 		if (tpnt->dynamic_info[DT_INIT]) {
-			_dl_elf_init = (int (*)(void)) (intptr_t) (tpnt->loadaddr + tpnt->dynamic_info[DT_INIT]);
-			  
+			void (*dl_elf_func) (void);
+			dl_elf_func = (void (*)(void)) (intptr_t) (tpnt->loadaddr + tpnt->dynamic_info[DT_INIT]);
 #if defined (__SUPPORT_LD_DEBUG__)
 			if(_dl_debug) _dl_dprintf(_dl_debug_file,"\ncalling init: %s\n\n", tpnt->libname);	
 #endif    
-			(*_dl_elf_init) ();
+			(*dl_elf_func) ();
 		}
 		if (_dl_atexit && tpnt->dynamic_info[DT_FINI]) {
-			(*_dl_atexit) (tpnt->loadaddr + tpnt->dynamic_info[DT_FINI]);
+			void (*dl_elf_func) (void);
+			dl_elf_func = (void (*)(void)) (intptr_t) (tpnt->loadaddr + tpnt->dynamic_info[DT_FINI]);
+			(*_dl_atexit) (dl_elf_func);
 #if defined (__SUPPORT_LD_DEBUG__)
 			if(_dl_debug && _dl_on_exit)
 			{
@@ -1306,62 +1306,6 @@ void _dl_debug_state(void)
 {
 }
 
-int _dl_fixup(struct elf_resolve *tpnt)
-{
-	int goof = 0;
-
-	if (tpnt->next)
-		goof += _dl_fixup(tpnt->next);
-#if defined (__SUPPORT_LD_DEBUG__)
-	if(_dl_debug) _dl_dprintf(_dl_debug_file,"\nrelocation processing: %s", tpnt->libname);	
-#endif    
-	
-	if (tpnt->dynamic_info[DT_REL]) {
-#ifdef ELF_USES_RELOCA
-		_dl_dprintf(2, "%s: can't handle REL relocation records\n",
-					_dl_progname);
-		_dl_exit(17);
-#else
-		if (tpnt->init_flag & RELOCS_DONE)
-			return goof;
-		tpnt->init_flag |= RELOCS_DONE;
-		goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_REL], 
-				tpnt->dynamic_info[DT_RELSZ], 0);
-#endif
-	}
-	if (tpnt->dynamic_info[DT_RELA]) {
-#ifdef ELF_USES_RELOCA
-		if (tpnt->init_flag & RELOCS_DONE)
-			return goof;
-		tpnt->init_flag |= RELOCS_DONE;
-		goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_RELA], 
-				tpnt->dynamic_info[DT_RELASZ], 0);
-#else
-		_dl_dprintf(2, "%s: can't handle RELA relocation records\n",
-					_dl_progname);
-		_dl_exit(18);
-#endif
-	}
-	if (tpnt->dynamic_info[DT_JMPREL]) {
-		if (tpnt->init_flag & JMP_RELOCS_DONE)
-			return goof;
-		tpnt->init_flag |= JMP_RELOCS_DONE;
-		if (!_dl_not_lazy || *_dl_not_lazy == 0)
-			_dl_parse_lazy_relocation_information(tpnt, tpnt->dynamic_info[DT_JMPREL], 
-					tpnt->dynamic_info [DT_PLTRELSZ], 0);
-		else
-			goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_JMPREL], 
-					tpnt->dynamic_info[DT_PLTRELSZ], 0);
-	}
-#if defined (__SUPPORT_LD_DEBUG__)
-	if(_dl_debug) {
-		_dl_dprintf(_dl_debug_file,"\nrelocation processing: %s", tpnt->libname);     
-		_dl_dprintf(_dl_debug_file,"; finished\n\n");
-	}
-#endif    
-	return goof;
-}
-
 char *_dl_getenv(const char *symbol, char **envp)
 {
 	char *pnt;
