@@ -21,6 +21,8 @@
 /* Make uClibc lie about being glibc. */
 #define __FORCE_GLIBC 1
 
+#include <locale.h>
+
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
@@ -463,10 +465,11 @@ _obstack_memory_used (h)
 
 /* Define the error handler.  */
 # ifndef _
-#  if (HAVE_LIBINTL_H && ENABLE_NLS) || defined _LIBC
+/* #  if (HAVE_LIBINTL_H && ENABLE_NLS) || defined _LIBC */
+#  ifdef __UCLIBC_HAS_GETTEXT_AWARENESS__
 #   include <libintl.h>
 #   ifndef _
-#    define _(Str) gettext (Str)
+#    define _(Str) __dcgettext (NULL, Str, LC_MESSAGES)
 #   endif
 #  else
 #   define _(Str) (Str)
